﻿//// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
//// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
//// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//// PARTICULAR PURPOSE.
////
//// Copyright (c) Microsoft Corporation. All rights reserved

#pragma once

#include <ppltasks.h> // For create_task
#include <iostream>
#include <iomanip>
#include <vector>
#include <Windows.h>
#include <wrl.h>
#include <d3d11_2.h>
#include <d2d1_2.h>
#include <dwrite_2.h>
#include <wincodec.h>
#include <d3dcompiler.h>
#include <DirectXMath.h>
#include <ppltasks.h>
#include <DirectXColors.h>
#include <memory>
#include <vector>
#include <deque>
#include <stack>
#include <algorithm>
#include <map>
#include <queue>
#include <list>
#include <sstream>
#include <Xinput.h>

typedef unsigned int uint32;
typedef unsigned long long uint64;


// Controls all the DirectX device resources.
class DeviceResources
{
public:
    DeviceResources();
    void SetWindow(HWND window);
    void UpdateForWindowSizeChange();
    void Present();

    // Device Accessors.
    RECT                    GetWindowBounds() const                 { return m_windowBounds; }

    // D3D Accessors.
    ID3D11Device2*          GetD3DDevice() const                    { return m_d3dDevice.Get(); }
    ID3D11DeviceContext2*   GetD3DDeviceContext() const             { return m_d3dContext.Get(); }
    IDXGISwapChain1*        GetSwapChain() const                    { return m_swapChain.Get(); }
    D3D_FEATURE_LEVEL       GetDeviceFeatureLevel() const           { return m_d3dFeatureLevel; }
    ID3D11RenderTargetView* GetBackBufferRenderTargetView() const   { return m_d3dRenderTargetView.Get(); }
    ID3D11DepthStencilView* GetDepthStencilView() const             { return m_d3dDepthStencilView.Get(); }
    D3D11_VIEWPORT          GetScreenViewport() const               { return m_screenViewport; }

    // D2D Accessors.
    ID2D1Factory2*          GetD2DFactory() const                   { return m_d2dFactory.Get(); }
    ID2D1Device1*           GetD2DDevice() const                    { return m_d2dDevice.Get(); }
    ID2D1DeviceContext1*    GetD2DDeviceContext() const             { return m_d2dContext.Get(); }
    ID2D1Bitmap1*           GetD2DTargetBitmap() const              { return m_d2dTargetBitmap.Get(); }
    IDWriteFactory2*        GetDWriteFactory() const                { return m_dwriteFactory.Get(); }
    IWICImagingFactory2*    GetWicImagingFactory() const            { return m_wicFactory.Get(); }

    // Sample-specific Accessors.
    D3D11_TILED_RESOURCES_TIER GetTiledResourcesTier() const        { return m_tiledResourcesTier; }

private:
    void CreateDeviceIndependentResources();
    void CreateDeviceResources();
    void CreateWindowSizeDependentResources();

    DXGI_MODE_ROTATION ComputeDisplayRotation();

    // Direct3D objects.
    Microsoft::WRL::ComPtr<ID3D11Device2>        m_d3dDevice;
    Microsoft::WRL::ComPtr<ID3D11DeviceContext2> m_d3dContext;
    Microsoft::WRL::ComPtr<IDXGISwapChain1>      m_swapChain;
    Microsoft::WRL::ComPtr<IDXGIFactory2> m_dxgiFactory;

    // Direct3D rendering objects. Required for 3D.
    Microsoft::WRL::ComPtr<ID3D11RenderTargetView> m_d3dRenderTargetView;
    Microsoft::WRL::ComPtr<ID3D11DepthStencilView> m_d3dDepthStencilView;
    D3D11_VIEWPORT                                 m_screenViewport;

    // Direct2D drawing components.
    Microsoft::WRL::ComPtr<ID2D1Factory2>       m_d2dFactory;
    Microsoft::WRL::ComPtr<ID2D1Device1>        m_d2dDevice;
    Microsoft::WRL::ComPtr<ID2D1DeviceContext1> m_d2dContext;
    Microsoft::WRL::ComPtr<ID2D1Bitmap1>        m_d2dTargetBitmap;

    // DirectWrite drawing components.
    Microsoft::WRL::ComPtr<IDWriteFactory2>     m_dwriteFactory;
    Microsoft::WRL::ComPtr<IWICImagingFactory2> m_wicFactory;

    // Cached reference to the Window.
    HWND m_window;

    // Cached device properties.
    D3D_FEATURE_LEVEL m_d3dFeatureLevel;
    SIZE              m_d3dRenderTargetSize;
    RECT              m_windowBounds;

    // Tiled Resources Tier.
    D3D11_TILED_RESOURCES_TIER m_tiledResourcesTier;
};


namespace DX
{
    inline void ThrowIfFailed(HRESULT hr)
    {
        if (FAILED(hr))
        {
            // Set a breakpoint on this line to catch Win32 API errors.
            std::wcout << L"FAILED HRESULT (0x" << std::hex << hr << L")" << std::endl;
            RaiseException(hr, 0, 0, NULL);
            exit(hr);
        }
    }

    // Function that reads from a binary file asynchronously.
    inline Concurrency::task<std::vector<byte>> ReadDataAsync(const std::wstring& filename)
    {
        using namespace Concurrency;

        HANDLE file = CreateFile2(
            filename.c_str(),
            GENERIC_READ,
            FILE_SHARE_READ,
            OPEN_EXISTING,
            NULL
            );

        DX::ThrowIfFailed(file != INVALID_HANDLE_VALUE ? S_OK : HRESULT_FROM_WIN32(GetLastError()));

        return create_task([file](){
            BY_HANDLE_FILE_INFORMATION info;
            if (!GetFileInformationByHandle(file, &info)) DX::ThrowIfFailed(HRESULT_FROM_WIN32(GetLastError()));

            std::vector<byte> ret(info.nFileSizeLow);
            DWORD read = 0;
            DX::ThrowIfFailed(ReadFile(file, ret.data(), info.nFileSizeLow, &read, NULL) ? S_OK : HRESULT_FROM_WIN32(GetLastError()));
            DX::ThrowIfFailed(read == info.nFileSizeLow ? S_OK : E_FAIL);

            CloseHandle(file);

            return ret;
        });
    }

    inline std::vector<byte> ReadData(const std::wstring& filename)
    {
        HANDLE file = CreateFile2(
            filename.c_str(),
            GENERIC_READ,
            FILE_SHARE_READ,
            OPEN_EXISTING,
            NULL
            );

        DX::ThrowIfFailed(file != INVALID_HANDLE_VALUE ? S_OK : HRESULT_FROM_WIN32(GetLastError()));

        BY_HANDLE_FILE_INFORMATION info;
        if (!GetFileInformationByHandle(file, &info)) DX::ThrowIfFailed(HRESULT_FROM_WIN32(GetLastError()));

        std::vector<byte> ret(info.nFileSizeLow);
        DWORD read = 0;
        DX::ThrowIfFailed(ReadFile(file, ret.data(), info.nFileSizeLow, &read, NULL) ? S_OK : HRESULT_FROM_WIN32(GetLastError()));
        DX::ThrowIfFailed(read == info.nFileSizeLow ? S_OK : E_FAIL);

        CloseHandle(file);

        return ret;
    }

#if defined(_DEBUG)
    // Check for SDK Layer support.
    inline bool SdkLayersAvailable()
    {
        HRESULT hr = D3D11CreateDevice(
            nullptr,
            D3D_DRIVER_TYPE_NULL,       // There is no need to create a real hardware device.
            0,
            D3D11_CREATE_DEVICE_DEBUG,  // Check for the SDK layers.
            nullptr,                    // Any feature level will do.
            0,
            D3D11_SDK_VERSION,          // Always set this to D3D11_SDK_VERSION for Windows Store apps.
            nullptr,                    // No need to keep the D3D device reference.
            nullptr,                    // No need to know the feature level.
            nullptr                     // No need to keep the D3D device context reference.
            );

        return SUCCEEDED(hr);
    }
#endif
}

class HR
{
public:
    HR(HRESULT hr)
    {
        m_hr = hr;
    }
    HRESULT operator=(HRESULT hr)
    {
        m_hr = hr;
        if (FAILED(hr))
        {
            std::wcout << L"FAILED HRESULT (0x" << std::hex << hr << L")" << std::endl;
            RaiseException(hr, 0, 0, NULL);
            exit(hr);
        }
        return hr;
    }
private:
    HRESULT m_hr;
};



void HandleXInput();

HWND MakeWindow();

void RunLoop( DeviceResources* pDevice );